import numpy as np
import warnings
import numba
import scipy.sparse as sp
import itertools
import numpy.ma as ma
import time


class Timer:
    def __init__(self, name):
        self._name = name

    def __enter__(self):
        self._start = time.time()
        return self

    def __exit__(self, *args):
        end = time.time()
        interval = end - self._start
        print(f"{self._name} took {interval:.2f} s")


class InvalidValuesWarning(UserWarning):
    pass


class Labels(object):
    """
    Abstract class for ground truth labels.
    """
    pass


@numba.njit(parallel=True)
def fast_lookup(A_indptr, A_cols, A_data, B):
    """
    Numba accelerated version of lookup table
    """
    # Non-existing indices are assigned label of 0.0
    vals = np.zeros(B.shape, dtype=np.float32)

    n_rows_a = len(A_indptr) - 1
    if n_rows_a == len(B):
        for i in numba.prange(B.shape[0]):
            ind_start, ind_end = A_indptr[i], A_indptr[i+1]
            left_idx = np.searchsorted(A_cols[ind_start:ind_end], B[i])
            right_idx = np.searchsorted(A_cols[ind_start:ind_end], B[i], side='right')
            found = (left_idx != right_idx)
            vals[i][found] = A_data[ind_start:ind_end][left_idx[found]]
    else:
        for i in numba.prange(B.shape[0]):
            left_idx = np.searchsorted(A_cols, B[i])
            right_idx = np.searchsorted(A_cols, B[i], side='right')
            found = (left_idx != right_idx)
            vals[i][found] = A_data[left_idx[found]]
    return vals


class BinaryLabels(Labels):
    """
    Represents binary ground truth data (e.g., 1 indicating relevance).
    """
    binary = True

    @classmethod
    def from_positive_indices(cls, indices):
        """
        Construct a binary labels instance from sparse data where only positive items are specified.

        Parameters
        ----------
        indices : array_like, one row per context (e.g., user or query)
                Specifies positive indices for each sample. Must be 1D or 2D, but row lengths can differ.

        Raises
        ------
        ValueError
                if `indices` is of invalid shape, type or contains duplicate, negative or non-integer indices.

        Examples
        --------
        >>> BinaryLabels.from_positive_indices([[1,2], [2]]) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        <rankereval.data.BinaryLabels...>
        """
        sp_matrix = cls._check_values(SparseMatrix.from_nonzero_indices(indices).tocsr(), binary=True)
        return cls()._init(sp_matrix)

    @classmethod
    def from_matrix(cls, labels):
        """
        Construct a binary labels instance from dense or sparse matrix where each item's label is specified.

        Parameters
        ----------
        labels : 1D or 2D array, one row per context (e.g., user or query)
                Contains binary labels for each item. Labels must be in {0, 1}.

        Raises
        ------
        ValueError
                if `labels` is of invalid shape, type or non-binary.

        Examples
        --------
        >>> BinaryLabels.from_matrix([[0, 1, 1], [0, 0, 1]]) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        <rankereval.data.BinaryLabels...>
        """
        sp_matrix = cls._check_values(SparseMatrix.from_matrix(labels).tocsr(), binary=cls.binary)
        return cls()._init(sp_matrix)

    def get_labels_for(self, ranking, k=None):
        n_label_rows = self._labels.shape[0]
        n_ranking_rows = len(ranking)

        if n_ranking_rows < n_label_rows:
            raise ValueError(
                f"Gold labels contain {n_label_rows} rows, but input rankings only have {n_ranking_rows} rows")

        indices, mask = ranking.get_top_k(k)
        retrieved = fast_lookup(self._labels.indptr,
                                self._labels.indices,
                                self._labels.data,
                                indices)
        return ma.masked_array(retrieved, mask=mask)

    @staticmethod
    @numba.njit
    def _numba_is_binary(data):
        for v in data:
            if v != 0 and v != 1:
                return False
        return True

    def as_rankings(self):
        return Rankings.from_scores(self._labels.tocsr(copy=True), warn_empty=False)

    def _init(self, labels):
        self._labels = labels
        return self

    @classmethod
    def _check_values(cls, matrix, binary=True):
        if binary and not cls._numba_is_binary(matrix.data):
            raise ValueError("Matrix may only contain 0 and 1 entries.")
        nonfinite_entries = ~np.isfinite(matrix.data)
        if np.any(nonfinite_entries):
            raise ValueError("Input contains NaN or Inf entries")
        return matrix

    def labels_to_list(self):
        return self._labels.tolil().data.tolist()

    def indices_to_list(self):
        return self._labels.tolil().rows.tolist()

    def get_n_positives(self, n_rankings):
        n_label_rows = self._labels.shape[0]
        n_pos = self._labels.getnnz(axis=1)
        if n_label_rows == 1:
            n_pos = np.tile(n_pos, n_rankings)
        return n_pos

    def __str__(self):
        return str(self.indices_to_list())


class NumericLabels(BinaryLabels):
    """
    Represents numeric ground truth data (e.g., relevance labels from 1-5).
    """
    binary = False


class SparseMatrix(object):
    """
    Stores sparse matrix data in unsorted CSR format (i.e., column indices in each row are unsorted).
    """

    def __init__(self, idx_ptr, col_idx, data, shape=None):
        self.idx_ptr = idx_ptr.copy()
        self.col_idx = col_idx.copy()
        self.data = data.copy()
        if shape:
            self.shape = shape
        else:
            if len(col_idx):
                M = col_idx.max() + 1
            else:
                M = 0
            self.shape = (len(idx_ptr) - 1, M)

    @classmethod
    def from_values(cls, data, keep_zeros=False):
        if isinstance(data, list):
            if len(data) == 0 or np.ndim(data[0]) == 0:
                data = [data]
            idx = [list(range(len(r))) for r in data]
            return cls.from_lil(idx, data, keep_zeros=keep_zeros)
        else:
            return cls.from_matrix(data, keep_zeros=keep_zeros)

    @classmethod
    def from_nonzero_indices(cls, indices):
        if sp.issparse(indices):
            x = indices.tocsr()
            return cls(x.indptr, x.indices, x.data, x.shape)
        else:
            return cls.from_lil(indices)

    @classmethod
    def from_matrix(cls, matrix, keep_zeros=False):
        if np.ma.isMaskedArray(matrix):
            raise ValueError("Masked arrays not supported.")
        elif isinstance(matrix, np.ndarray) or isinstance(matrix, list):
            if isinstance(matrix, list):
                matrix = np.asarray(matrix, dtype=object).astype(np.float32)
            matrix = np.atleast_2d(matrix)
            if not np.issubdtype(matrix.dtype, np.number) or np.issubdtype(matrix.dtype, np.bool_):
                raise ValueError("Input must be numeric")
            elif matrix.ndim != 2:
                raise ValueError("Input arrays need to be 1D or 2D.")
            if keep_zeros:
                matrix += 1 - matrix[np.isfinite(matrix)].min()
            x = sp.csr_matrix(matrix)
            if not keep_zeros:
                x.eliminate_zeros()
        elif sp.issparse(matrix):
            x = matrix.tocsr()
        else:
            raise ValueError("Input type not supported.")
        return cls(x.indptr, x.indices, x.data, x.shape)

    @classmethod
    def from_lil(cls, rows, data=None, dtype=np.float32, keep_zeros=False):
        if not isinstance(rows, list) and not isinstance(rows, np.ndarray):
            raise ValueError("Invalid input type.")
        if len(rows) == 0 or np.ndim(rows[0]) == 0:
            rows = [rows]
        idx_ptr = np.asarray([0] + [len(x) for x in rows], dtype=int).cumsum()
        try:
            col_idx = np.fromiter(itertools.chain.from_iterable(rows), dtype=int, count=idx_ptr[-1])
            if data is None:
                data = np.ones_like(col_idx, dtype=dtype)
            else:
                data = np.fromiter(itertools.chain.from_iterable(data), dtype=dtype, count=idx_ptr[-1])
                if keep_zeros:
                    data += 1 - data[np.isfinite(data)].min()
        except TypeError:
            raise ValueError("Invalid values in input.")
        if len(data) != len(col_idx):
            raise ValueError("rows and data need to have same length")
        instance = cls(idx_ptr, col_idx, data)
        if not keep_zeros:
            instance.eliminate_zeros()
        return instance

    def max_nnz_row_values(self):
        """Returns maximum number of non-zero entries in any row."""
        return (self.idx_ptr[1:] - self.idx_ptr[:-1]).max()

    def count_empty_rows(self):
        return ((self.idx_ptr[1:] - self.idx_ptr[:-1]) == 0).sum()

    def sort(self):
        self._numba_sort(self.idx_ptr, self.col_idx, self.data)

    def intersection(self, other):
        self._setop(other, True)

    def difference(self, other):
        self._setop(other, False)

    def isfinite(self):
        return np.all(np.isfinite(self.data))

    def remove_infinite(self):
        if not self.isfinite():
            self.data[~np.isfinite(self.data)] = 0
            self.eliminate_zeros()

    def eliminate_zeros(self):
        csr = self.tocsr()
        csr.eliminate_zeros()
        self.data, self.col_idx, self.idx_ptr = csr.data, csr.indices, csr.indptr

    def _setop(self, other, mode):
        if self.shape[0] != other.shape[0]:
            raise ValueError("Matrices need to have the same number of rows!")
        self._numba_setop(self.idx_ptr, self.col_idx, self.data, other.idx_ptr, other.col_idx, mode)
        self.eliminate_zeros()

    def tocsr(self):
        return sp.csr_matrix((self.data, self.col_idx, self.idx_ptr), copy=False, shape=self.shape)

    def tolil(self):
        res = []
        for i in range(len(self.idx_ptr) - 1):
            start, end = self.idx_ptr[i], self.idx_ptr[i+1]
            res += [self.col_idx[start:end].tolist()]
        return res

    def todense(self):
        return np.asarray(self.tocsr().todense())

    @staticmethod
    @numba.njit(parallel=True)
    def _numba_sort(idx_ptr, col_idx, data):
        for i in numba.prange(len(idx_ptr) - 1):
            start, end = idx_ptr[i], idx_ptr[i+1]
            args = (-data[start:end]).argsort(kind="mergesort")
            data[start:end] = data[start:end][args]
            col_idx[start:end] = col_idx[start:end][args]

    @staticmethod
    @numba.njit(parallel=True)
    def _numba_setop(self_idx_ptr, self_col_idx, self_data, other_idx_ptr, other_col_idx, intersect):
        for i in numba.prange(len(self_idx_ptr) - 1):
            ss, se = self_idx_ptr[i], self_idx_ptr[i+1]
            os, oe = other_idx_ptr[i], other_idx_ptr[i+1]

            left_idx = np.searchsorted(other_col_idx[os:oe], self_col_idx[ss:se])
            right_idx = np.searchsorted(other_col_idx[os:oe], self_col_idx[ss:se], side='right')
            if intersect:
                found = (left_idx == right_idx)
            else:
                found = (left_idx != right_idx)
            self_data[ss:se][found] = 0

    def __str__(self):
        return str((self.idx_ptr, self.col_idx, self.data))


class Rankings(object):
    """
    Represents (predicted) rankings to be evaluated.
    """

    def __init__(self, indices, valid_items=None, invalid_items=None, warn_empty=True):
        if valid_items is not None:
            valid_items = SparseMatrix.from_nonzero_indices(valid_items)
            indices.intersection(valid_items)
        if invalid_items is not None:
            invalid_items = SparseMatrix.from_nonzero_indices(invalid_items)
            indices.difference(invalid_items)
        if not indices.isfinite():
            warnings.warn("Input contains NaN or Inf entries which will be ignored.",
                          InvalidValuesWarning)
            indices.remove_infinite()
        n_empty_rows = indices.count_empty_rows()
        if n_empty_rows and warn_empty:
            warnings.warn(f"Input rankings have {n_empty_rows} empty rankings (rows). "
                          + "These will impact the mean scores." + str(indices.todense()),
                          InvalidValuesWarning)
        self.indices = indices

    @classmethod
    def from_ranked_indices(cls, indices, valid_items=None, invalid_items=None):
        """
        Construct a rankings instance from data where item indices are specified in ranked order.

        Parameters
        ----------
        indices : array_like, one row per ranking
                Indices of items after ranking. Must be 1D or 2D, but row lengths can differ.
        valid_items : array_like, one row per ranking
                Indices of valid items (e.g., candidate set). Invalid items will be discarded from ranking.

        Raises
        ------
        ValueError
                if `indices` or `valid_items` of invalid shape or type.

        Examples
        --------
        >>> Rankings.from_ranked_indices([[5, 2], [4, 3, 1]]) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        <rankereval.data.Rankings...>
        """
        indices = SparseMatrix.from_lil(indices)
        return cls(indices, valid_items, invalid_items)

    @classmethod
    def from_scores(cls, raw_scores, valid_items=None, invalid_items=None, warn_empty=True):
        """
        Construct a rankings instance from raw scores where each item's score is specified.
        Items will be ranked in descending order (higher scores meaning better).

        Parameters
        ----------
        raw_scores : array_like, one row per ranking
                Contains raw scores for each item. Must be 1D or 2D, but row lengths can differ.
        valid_items : array_like, one row per ranking
                Indices of valid items (e.g., candidate set). Invalid items will be discarded from ranking.

        Raises
        ------
        ValueError
                if `raw_scores` or `valid_items` of invalid shape or type.

        Warns
        ------
        InvalidValuesWarning
                if `raw_scores` contains non-finite values.

        Examples
        --------
        >>> Rankings.from_scores([[0.1, 0.5, 0.2], [0.4, 0.2, 0.5]]) # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
        <rankereval.data.Rankings...>
        """
        indices = SparseMatrix.from_values(raw_scores, keep_zeros=True)
        indices.sort()

        return cls(indices, valid_items, invalid_items, warn_empty=warn_empty)

    def __str__(self):
        return str(self.indices)

    def __len__(self):
        return self.indices.shape[0]

    def to_list(self):
        return self.indices.tolil()

    def get_top_k(self, k=None):
        if k is None:
            k = self.indices.max_nnz_row_values()
        return self._csr_to_dense_masked(self.indices.idx_ptr, self.indices.col_idx, (len(self), k))

    @staticmethod
    @numba.njit
    def _csr_to_dense_masked(idx_ptr, col_idx, shape):
        condensed = np.zeros(shape, dtype=col_idx.dtype)
        mask = np.ones(shape, dtype=np.bool_)
        for i in range(len(idx_ptr) - 1):
            start, end = idx_ptr[i], idx_ptr[i+1]
            length = min(end - start, shape[1])
            condensed[i][:length] = col_idx[start:start+length]
            mask[i][:length] = False
        return condensed, mask


class TopKMixin:
    @staticmethod
    def topk(x, k, return_scores=False):
        # partition into k largest elements first
        index_array = np.sort(np.argpartition(-x, kth=k-1, axis=-1)[:, :k])
        top_k_partition = np.take_along_axis(x, index_array, axis=-1)

        # stable argsort in descending order
        top_idx_local = top_k_partition.shape[1] - 1
        top_idx_local -= np.fliplr(np.argsort(np.fliplr(top_k_partition), axis=-1, kind='stable'))

        # sort the top partition
        top_idx = np.take_along_axis(index_array, top_idx_local, axis=-1)
        if not return_scores:
            return top_idx
        else:
            top_scores = np.take_along_axis(top_k_partition, top_idx_local, axis=-1)
            return top_scores, top_idx


class DenseRankings(Rankings, TopKMixin):
    """
    Data structure where rankings have the same length (approximately).
    """

    def __init__(self, indices, mask=None, warn_empty=True):
        n_empty_rows = ((~mask).sum(axis=1) == 0).sum()
        if n_empty_rows and warn_empty:
            warnings.warn(f"Input rankings have {n_empty_rows} empty rankings (rows). "
                          + "These will impact the mean scores." + str(indices.csr.todense()),
                          InvalidValuesWarning)
        self.indices = indices
        self.mask = mask

    @classmethod
    def _verify_input(cls, arr, dtype=np.floating):
        if not isinstance(arr, np.ndarray):
            raise ValueError("Input needs to be a numpy matrix")
        arr = np.asarray(np.atleast_2d(arr))
        if arr.ndim != 2:
            raise ValueError("Input arrays need to be 1D or 2D.")
        elif not np.issubdtype(arr.dtype, dtype):
            raise ValueError(f"Input array needs to be of type {dtype}")

        if np.issubdtype(dtype, np.floating):
            if not np.all(np.isfinite(arr)):
                warnings.warn("Input contains NaN or Inf entries which will be ignored.",
                              InvalidValuesWarning)
                arr[~np.isfinite(arr)] = np.NINF
        elif not np.issubdtype(dtype, np.integer):
            raise TypeError("dtype argument must be floating or int")
        return arr

    @classmethod
    def from_ranked_indices(cls, indices, valid_items=None, invalid_items=None):
        """
        Set indices to -1 (or any other negative value) to indicate invalid index
        """
        indices = cls._verify_input(indices, dtype=np.integer)

        if valid_items is not None or invalid_items is not None:
            raise NotImplementedError("Not implemented yet")
        mask = (indices < 0)
        return cls(indices, mask)

    @classmethod
    def from_scores(cls, raw_scores, valid_items=None, invalid_items=None, warn_empty=True, k_max=None):
        raw_scores, mask = cls._verify_input(raw_scores, dtype=np.floating)

        if valid_items is not None:
            invalid_idx = SparseMatrix.from_nonzero_indices(invalid_items).csr.toarray() == 0
            raw_scores -= np.inf*invalid_idx
        if invalid_items is not None:
            invalid_items = SparseMatrix.from_nonzero_indices(invalid_items).csr
            raw_scores -= np.inf*invalid_items

        mask = ~np.isfinite(raw_scores)
        if k_max is None:
            sorted_idx = np.argsort(-raw_scores, axis=1, kind="stable")
        else:
            sorted_idx = cls.topk(raw_scores, k_max)
        mask = np.take_along_axis(mask, sorted_idx, axis=1)
        return cls(sorted_idx, mask)

    def get_top_k(self, k=None):
        if k is None:
            k = self.indices.shape[1]
        indices = self.indices[:, :k]
        mask = self.mask[:, :k]
        return indices, mask

    def to_list(self):
        return self.indices.tolist()



import numpy as np
from .data import BinaryLabels, Labels, Rankings


class Metric(object):
    def score(self, y_true, y_pred):
        r"""
        Individual scores for each ranking.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.Labels`
                Ground truth labels.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.
        ValueError
                if `n_bootstrap_samples`, `confidence` or `nan_handling` contain invalid values.
        """
        if not isinstance(y_true, Labels):
            raise TypeError("y_true must be of type Labels")
        if not isinstance(y_pred, Rankings):
            raise TypeError("y_pred must be of type Rankings")

        y_pred_labels = y_true.get_labels_for(y_pred, self._k)

        return self._score(y_true, y_pred_labels)

    @classmethod
    def _bootstrap_ci(cls, scores, n_bootstrap_samples, confidence):
        if not isinstance(n_bootstrap_samples,
                          int) or n_bootstrap_samples <= 1:
            raise ValueError("n_bootstrap_samples must be int > 1")
        elif not isinstance(confidence, float) or confidence <= 0.0 or confidence >= 1.0:
            raise ValueError("Confidence must be float and 0 < confidence < 1")

        if len(scores):
            resamples = np.random.choice(
                scores, (len(scores), n_bootstrap_samples), replace=True)
            bootstrap_means = resamples.mean(axis=0)

            # Compute "percentile bootstrap"
            alpha_2 = (1 - confidence) / 2.0
            lower_ci = np.quantile(bootstrap_means, alpha_2)
            upper_ci = np.quantile(bootstrap_means, 1.0 - alpha_2)
            return (lower_ci, upper_ci)
        else:
            return (float('nan'), float('nan'))

    def mean(self, y_true, y_pred, nan_handling='propagate',
             conf_interval=False, n_bootstrap_samples=1000, confidence=0.95):
        r"""
        Mean score over all ranking after handling NaN values.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.Labels`
                Ground truth labels, see also above.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated.
        nan_handling : {'propagate', 'drop', 'zerofill'}, optional
                `'propagate'` (default):
                        Return NaN if any value is NaN
                `'drop'` :
                        Ignore NaN values
                `'zerofill'` :
                        Replace NaN values with zero
        conv_interval : bool, optional
                If True, then return bootstrapped confidence intervals of mean,
                otherwise interval is None.
                Defaults to False.
        n_bootstrap_samples : int, optional
                Number of bootstrap samples to draw.
        confidence : float, optional
                Indicates width of confidence interval. Default is 0.95 (95%).
        Returns
        -------
        mean: float `mean` if `conv_interval` is `false` otherwise
                Dictionary with ``mean["score"]`` and ``mean["conf_interval"]``
                for the confidence interval tuple `(lower CI, upper CI)`.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.
        ValueError
                if `n_bootstrap_samples`, `confidence` or `nan_handling` contain invalid
                values.
        """

        scores = self.score(y_true, y_pred)
        if nan_handling == 'drop':
            scores = scores[~np.isnan(scores)]
        elif nan_handling == 'zerofill':
            scores = np.nan_to_num(scores)
        elif nan_handling == "propagate":
            if np.isnan(scores).sum():
                scores = []
        else:
            raise ValueError(
                'nan_handling must be "propagate", "drop" or "zerofill"')

        if conf_interval:
            ci = self._bootstrap_ci(scores, n_bootstrap_samples, confidence)
        else:
            ci = None

        if len(scores):
            mean = scores.mean()
        else:
            mean = float('nan')

        if conf_interval:
            return {"score": mean, "conf_interval": ci}
        else:
            return mean

    def name(self):
        if self._k is None:
            k = ""
        else:
            k = f"@{self._k}"
        return self.__class__.__name__ + k


class BinaryMetric(Metric):
    def __init__(self, k):
        if not isinstance(k, int) or k <= 0:
            raise ValueError("Cutoff k needs to be integer > 0")
        self._k = k

    def score(self, y_true, y_pred):
        if not isinstance(y_true, BinaryLabels):
            raise TypeError(f"y_true must be of type BinaryLabels but is of instance {type(y_true)}")
        return super().score(y_true, y_pred)


class Precision(BinaryMetric):
    """
            Parameters
            ----------
            k : int
                    specifies number of top results `k` of each ranking to be evaluated.
            truncated : bool
                    if `true`, `k` gets clipped at length of input ranking.

            Raises
            ------
            ValueError
                    if `k` is not integer > 0.
    """

    def __init__(self, k, truncated=False):
        super().__init__(k)
        self._truncated = truncated

    def _precision(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        n_relevant = np.sum(y_pred_labels[:, :self._k] == 1, axis=-1).filled(0)
        if self._truncated:
            items = np.broadcast_to(~y_pred_labels.mask, y_pred_labels.shape)
            n_items_in_y_pred = items.sum(axis=1).flatten()

            # not defined if there are no relevant labels
            scores = np.NaN * np.zeros_like(n_relevant, dtype=float)
            valid = (n_items_in_y_pred > 0) & (n_pos > 0)

            scores[valid] = n_relevant[valid].astype(float) / np.minimum(n_items_in_y_pred[valid], self._k)
        else:
            scores = n_relevant.astype(float) / self._k
            # not defined if there are no relevant labels
            scores[n_pos == 0] = np.NaN
        return scores

    def _score(self, y_true, y_pred_labels):
        return self._precision(y_true, y_pred_labels)

    def score(self, y_true, y_pred):
        r"""
        Computes Precision@k [MN]_ of each ranking *y* in `y_pred` as

        .. math::

                \mathrm{Precision}@k(y) = \frac{\sum_{i=1}^{k'} \mathrm{rel}(y_i)}{k'},

        where :math:`\mathrm{rel}(y_i)` is the relevance label of the item at rank *i*
        in the ranking *y*. If `truncated` then :math:`k' = \min(k,|y|)`, i.e., `k'`
        can never exceed the length of the input ranking; otherwise :math:`k' = k` (default).

        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. Ranking to be evaluated is empty, i.e., :math:`|y|=0`. Per definition above,
        :math:`\mathrm{Precision}@k(y) = 0`.

        2. There are no relevant items in `y_true`: :math:`\mathrm{Precision}@k(y) =` NaN.
        This marks invalid instances explicitly and is consistent with Recall.

        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, Precision
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0, 5],[1,2,3]])
        >>> y_pred = Rankings.from_ranked_indices([[3,2,1], [1,2]])
        >>> Precision(3).score(y_true, y_pred)
        array([0.        , 0.66666667])

        """
        return super().score(y_true, y_pred)


class Recall(BinaryMetric):
    """
            Parameters
            ----------
            k : int
                    specifies number of top results `k` of each ranking to be evaluated.
            truncated : bool
                    if `true`, number of relevant results gets clipped at `k`.
            Raises
            ------
            ValueError
                    if `k` is not integer > 0.
    """

    def __init__(self, k, truncated=False):
        super().__init__(k)
        self._truncated = truncated

    def _recall(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        n_relevant = np.sum(y_pred_labels[:, :self._k] == 1, axis=-1).filled(0)

        scores = np.NaN * np.zeros_like(n_relevant, dtype=float)
        if self._truncated:
            denominator = np.clip(n_pos[n_pos > 0], None, self._k)
        else:
            denominator = n_pos[n_pos > 0]
        scores[n_pos > 0] = n_relevant[n_pos > 0].astype(float) / denominator
        return scores

    def _score(self, y_true, y_pred_labels):
        return self._recall(y_true, y_pred_labels)

    def score(self, y_true, y_pred):
        r"""
        Computes Recall@k [MN]_ as the fraction of relevant results in `y_true` that were
        in the top *k* results of `y_pred`.
        More formally, the recall of each ranking *y* in `y_pred` with
        labels `y_true` is defined as

        .. math::

                \mathrm{Recall}@k(y) =
                \frac{\sum_{i=1}^{k'} \mathrm{rel}(y_i)}{n_{rel}},

        where :math:`\mathrm{rel}(y_i)` is the relevance label of the item at rank *i*
        in the ranking *y* and :math:`n_{rel} = \left\| y_{true} \right\| _1` if `truncated`
        is false (default) and :math:`n_{rel} = \max(\left\| y_{true} \right\| _1, k)` otherwise.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. Ranking to be evaluated is empty, i.e., :math:`|y|=0`. Per definition above,
        :math:`\mathrm{Recall}@k(y) = 0`.

        2. There are no relevant items in `y_true`: :math:`\mathrm{Recall}@k(y) =` NaN. This
        marks invalid instances explicitly.

        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, Recall
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0, 5],[1]])
        >>> y_pred = Rankings.from_ranked_indices([[3,2,1], [1,2]])
        >>> Recall(3).score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([0., 1.])

        """
        return super().score(y_true, y_pred)


class F1(Precision, Recall):
    def score(self, y_true, y_pred):
        r"""
        Computes F1 [MN]_ as harmonic mean of Precision@k and Recall@k.
        More formally, the F1 score of each ranking *y* in `y_pred` is defined as

        .. math::

                \mathrm{F1}@k(y) = \frac{2*\big(\mathrm{Precision}@k(y) *
                \mathrm{Recall}@k(y)\big)}{\mathrm{Precision}@k(y) + \mathrm{Recall}@k(y)}.


        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. Ranking to be evaluated is empty, i.e., :math:`|y|=0`. In this case,
        :math:`\mathrm{Recall}@k(y) = \mathrm{Precision}@k(y) = 0` .

        2. If :math:`\mathrm{Recall}@k(y) = \mathrm{Precision}@k(y) = 0`,
        we define :math:`\mathrm{F1}@k(y) = 0`.

        3. There are no relevant items in `y_true`: :math:`\mathrm{F1}@k(y) =` NaN.

        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, F1
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0, 5],[1]])
        >>> y_pred = Rankings.from_ranked_indices([[3,2,1], [1,2]])
        >>> F1(3).score(y_true, y_pred)  # doctest: +NORMALIZE_WHITESPACE
        array([0. , 0.5])

        """
        return super().score(y_true, y_pred)

    def _score(self, y_true, y_pred_labels):
        recall = self._recall(y_true, y_pred_labels)
        precision = self._precision(y_true, y_pred_labels)

        product = 2 * recall * precision
        sm = recall + precision

        # return 0 for geometric mean if both are zero
        scores = np.zeros_like(product, dtype=float)
        valid = np.nan_to_num(product) > 0
        invalid = np.isnan(product)

        scores[valid] = product[valid] / sm[valid]
        scores[invalid] = np.NaN

        return scores


class HitRate(Recall):
    """
            Parameters
            ----------
            k : int
                    specifies number of top results `k` of each ranking to be evaluated.

            Raises
            ------
            ValueError
                    if `k` is not integer > 0.
    """

    def score(self, y_true, y_pred):
        r"""
        Computes HitRate@k [MN]_ as the whether a relevant item occurs
        in *k* results of `y_pred`.
        Differs from Recall@k in that `y_true` has to contain exactly
        one element per row.
        More formally, the HitRate of each ranking *y* in `y_pred` with labels
        `y_true` is defined as

        .. math::

                \mathrm{HitRate}@k(y) &= \sum_{i=1}^{k'} \mathrm{rel}(y_i),

                k' &= \min(k,|y|);

        where :math:`\mathrm{rel}(y_i)` is the relevance label of the item at rank *i*
        in the ranking *y*.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary and exactly one relevant item per row.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. Ranking to be evaluated is empty, i.e., :math:`|y|=0`.
        Per definition above, :math:`\mathrm{HitRate}@k(y) = 0`.

        2. There is not exactly one relevant item in
        `y_true`: :math:`\mathrm{HitRate}@k(y) =` NaN.


        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, HitRate
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0],[1]])
        >>> y_pred = Rankings.from_ranked_indices([[3,2,1], [1,2]])
        >>> HitRate(3).score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([0., 1.])

        """
        return super().score(y_true, y_pred)

    def _score(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        scores = self._recall(y_true, y_pred_labels)
        scores[n_pos != 1] = np.NaN  # Not defined for no or multiple positives
        return scores


class ReciprocalRank(BinaryMetric):
    """
            Parameters
            ----------
            k : int
                    specifies number of top results `k` of each ranking to be evaluated.

            Raises
            ------
            ValueError
                    if `k` is not integer > 0.
    """

    def _score(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        labels = y_pred_labels[:, :self._k].filled(0)
        ranks = np.arange(1, labels.shape[1] + 1, dtype=float).reshape(1, -1)

        # It is 1/rank if document appears in top k, 0 otherwise
        scores = np.max(labels / ranks, axis=-1, initial=0.0)
        scores[n_pos == 0] = np.NaN  # Not defined for no multiple positives

        return scores

    def score(self, y_true, y_pred):
        r"""
        Computes ReciprocalRank@k [NC]_ as the rank where the first relevant item
        occurs in the top *k* results of `y_pred`.
        More formally, the ReciprocalRank of each ranking *y* in `y_pred` is defined as

        .. math::

                \mathrm{ReciprocalRank}@k(y) &= \max_{i=1,\ldots,k'} \frac{\mathrm{rel}(y_i)}{i},

                k' &= \min(k,|y|);

        where :math:`\mathrm{rel}(y_i)` is the relevance label of the item at rank *i*
        in the ranking *y*.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. Ranking to be evaluated is empty or first relevant item appears beyond rank *k*.
        Per definition above, :math:`\mathrm{ReciprocalRank}@k(y) = 0`.

        2. There is no relevant item in `y_true`: :math:`\mathrm{ReciprocalRank}@k(y) =` NaN.


        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, ReciprocalRank
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0, 5],[1]])
        >>> y_pred = Rankings.from_ranked_indices([[3,0,1], [1,2]])
        >>> ReciprocalRank(3).score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([0.5, 1. ])

        """
        return super().score(y_true, y_pred)


class MeanRanks(BinaryMetric):
    """
            Used for evaluating permutations of `y_true`. Does not accept *k* as it
            scores permutations.
    """

    def __init__(self):
        self._k = None

    def _score(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        labels = y_pred_labels.filled(0)
        ranks = np.arange(1, labels.shape[1] + 1, dtype=float).reshape(1, -1)

        scores = np.sum(ranks * labels, axis=-1)
        scores[n_pos > 0] = scores[n_pos > 0] / n_pos[n_pos > 0]
        scores[n_pos == 0] = np.NaN
        return scores

    def score(self, y_true, y_pred):
        r"""
        Computes MeanRanks as the mean of ranks of relevant items `y_pred`.
        More formally, it is defined for each ranking *y* in `y_pred` as

        .. math::

                \mathrm{MeanRanks}(y) =
                \frac{\sum_{i=1}^{|y|} i\cdot\mathrm{rel}(y_i)}{\sum_{i=1}^{|y|} \mathrm{rel}(y_i)},

        where :math:`\mathrm{rel}(y_i)` is the relevance label of the item at rank *i*
        in the ranking *y*.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. There is no relevant item in `y_true`: :math:`\mathrm{MeanRanks}(y) =` NaN.


        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, MeanRanks
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0, 5],[1]])
        >>> y_pred = Rankings.from_ranked_indices([[3,0,5], [1,2]])
        >>> MeanRanks().score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([2.5, 1. ])

        """
        return super().score(y_true, y_pred)


class FirstRelevantRank(BinaryMetric):
    """
            Used for evaluating permutations of `y_true`. Does not accept *k* as it
            scores permutations.
    """

    def __init__(self):
        self._k = None

    def _score(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        labels = y_pred_labels.filled(0)
        ranks = np.arange(1, labels.shape[1] + 1, dtype=float).reshape(1, -1)

        ranks = ranks * labels
        ranks[ranks == 0] = np.inf
        scores = np.min(ranks, axis=-1)
        scores[n_pos == 0] = np.NaN
        return scores

    def score(self, y_true, y_pred):
        r"""
        Computes FirstRelevantRank as the mean of ranks of relevant items `y_pred`.
        More formally, it is defined for each ranking *y* in `y_pred` as

        .. math::

                \mathrm{FirstRelevantRank}(y) = \min \{i \mid \mathrm{rel}(y_i) = 1\},

        where :math:`\mathrm{rel}(y_i)` is the relevance label of the item at rank *i*
        in the ranking *y*.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. There is no relevant item in `y_true`: :math:`\mathrm{FirstRelevantRank}(y) =` NaN.


        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, FirstRelevantRank
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0, 5],[1]])
        >>> y_pred = Rankings.from_ranked_indices([[3,0,5], [1,2]])
        >>> FirstRelevantRank().score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([2., 1. ])

        """
        return super().score(y_true, y_pred)


class AP(BinaryMetric):
    """
            Parameters
            ----------
            k : int
                    specifies number of top results `k` of each ranking to be evaluated.

            Raises
            ------
            ValueError
                    if `k` is not integer > 0.
    """

    def _score(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        labels = y_pred_labels[:, :self._k].filled(0)
        ranks = np.arange(1, labels.shape[1] + 1, dtype=float).reshape(1, -1)

        precision = np.cumsum(labels, axis=-1) / ranks

        scores = np.zeros_like(n_pos, dtype=float)
        scores[n_pos > 0] = np.sum(
            precision * labels, axis=-1)[n_pos > 0] / np.clip(n_pos[n_pos > 0], None, self._k)
        scores[n_pos == 0] = np.NaN

        return scores

    def score(self, y_true, y_pred):
        r"""
        Computes AveragePrecision@k [MN]_, an approximation to the the area
        under the precision-recall curve, of each ranking *y* in `y_pred` as

        .. math::

                \mathrm{AveragePrecision}@k(y) &= \frac{1}{Z} \sum_{i=1}^{k'}
                \mathrm{rel}(y_i)\cdot \mathrm{Precision}@i(y),

                k' &= \min(k,|y|),

                Z &= \min \big(k, \left\| y_{true} \right\| _1\big);

        where :math:`\mathrm{rel}(y_i)` is the relevance label of the item at
        rank *i* in the ranking *y*.

        .. note::

                Sometimes the denominator *Z* is defined with respect to only the
                retrieved or recommended set of items `y_pred`. This is not desirable
                as AP could be artificially inflated, e.g., by returning only one
                relevant item at the top
                and then filling up the ranking with and irrelevant items.


        Parameters
        ----------
        y_true : :class:`~rankereval.data.BinaryLabels`
                Ground truth labels, must be binary.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. Ranking to be evaluated is empty, i.e., :math:`|y|=0`.
        Per definition above, :math:`\mathrm{AveragePrecision}@k(y) = 0`.

        2. There are no relevant items in `y_true`:
        :math:`\mathrm{AveragePrecision}@k(y) =` NaN to make it consistent with other metrics.

        3. There are no relevant items in `y_pred` up to *k*:
        Per definition above, :math:`\mathrm{AveragePrecision}@k(y) = 0`.


        Examples
        --------
        >>> from rankereval import BinaryLabels, Rankings, AP
        >>> # use separate labels for each ranking
        >>> y_true = BinaryLabels.from_positive_indices([[0, 5],[1,2,3]])
        >>> y_pred = Rankings.from_ranked_indices([[3,2,1], [1,2]])
        >>> AP(3).score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([0. , 0.66666667])

        """
        return super().score(y_true, y_pred)


class DCG(Metric):
    r"""
            Parameters
            ----------
            k : int
                    specifies number of top results `k` of each ranking to be evaluated.

            relevance_scaling : str, ['binary' (default), 'power']
                    Determines are relevance labels are transformed:

                    `'identity'`: (default)
                            :math:`f(\mathrm{rel}(y_i)) = \mathrm{rel}(y_i)`
                    `'power'`:
                            :math:`f(\mathrm{rel}(y_i)) = 2^{\mathrm{rel}(y_i)} - 1`

            log_base : str, ['e' (default), '2']
                    Determines what log base is used in denominator.
                    The smaller this value, the heavier emphasis on top-ranked documents.

                    `'e'` (default):
                            Natural logarithm :math:`\ln`
                    `'2'`:
                            :math:`\log_2`

            Notes
            -----
            The original definition of (n)DCG [KJ]_ uses 'identity' for `relevance_scaling`,
            but leaves the choice of `log_base` open.

            Raises
            ------
            ValueError
                    if `k` is not integer > 0 or `relevance_scaling` or `log_base` are invalid.

    """
    SCALERS = {'identity': lambda x: x,
               'power': lambda x: np.power(x, 2) - 1}
    LOGS = {'2': lambda x: np.log2(x), 'e': lambda x: np.log(x)}

    def __init__(self, k=None, relevance_scaling='identity', log_base='e'):
        self._k = k
        if relevance_scaling not in self.SCALERS:
            raise ValueError(
                "Relevance scaling must be 'identity' or 'power'.")
        if log_base not in self.LOGS:
            raise ValueError("Log base needs to be 'e' or '2'.")
        self._rel_scale = self.SCALERS[relevance_scaling]
        self._log_fct = self.LOGS[log_base]

    def _dcg(self, y_true, y_pred_labels):
        n_pos = y_true.get_n_positives(y_pred_labels.shape[0])
        labels = y_pred_labels[:, :self._k].filled(0)
        ranks = np.arange(1, labels.shape[1] + 1, dtype=float).reshape(1, -1)

        scores = np.sum(
            self._rel_scale(labels) /
            self._log_fct(
                ranks +
                1),
            axis=-
            1)
        scores[n_pos == 0] = np.NaN
        return scores

    def _score(self, y_true, y_pred_labels):
        return self._dcg(y_true, y_pred_labels)

    def score(self, y_true, y_pred):
        r"""
        Computes Discounted Cumulative Gain at k (DCG@k) [KJ]_ as a
        weighted sum of relevance labels of top *k* results of `y_pred`.
        More formally, the recall of each ranking *y* in `y_pred`
        with labels `y_true` is defined as

        .. math::

                \mathrm{DCG}@k(y) &= \sum_{i=1}^{k'} \frac{f(\mathrm{rel}(y_i))}{\log_b(i + 1)},

                k' &= \min(k,|y|);

        where :math:`\mathrm{rel}(y_i)` is the relevance label of
        the item at rank *i* in the ranking *y*.
        *f* is the `relevance_scaling` function and *b* the `log_base`
        parameters defined earlier.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.Labels`
                Ground truth labels, binary or numeric.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. Ranking to be evaluated is empty, i.e., :math:`|y|=0`.
        Per definition above, :math:`\mathrm{DCG}@k(y) = 0`.

        2. There are no items with relevance > 0 in `y_true`:
        :math:`\mathrm{DCG}@k(y) =` NaN to make it consistent with other metrics.

        Examples
        --------
        >>> from rankereval import NumericLabels, Rankings, DCG
        >>> # use separate labels for each ranking
        >>> y_true = NumericLabels.from_dense([[1, 2, 3], [4, 5]])
        >>> y_pred = Rankings.from_ranked_indices([[0,2,1], [1,0]])
        >>> DCG(3).score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([ 5.61610776, 10.85443211])

        """
        return super().score(y_true, y_pred)


class NDCG(DCG):
    """
    For a description of the arguments, see :class:`DCG`.
    """

    def _score(self, y_true, y_pred_labels):
        dcg = self._dcg(y_true, y_pred_labels)
        ideal_labels = y_true.get_labels_for(y_true.as_rankings(), self._k)
        idcg = self._dcg(y_true, ideal_labels)

        return dcg / idcg

    def score(self, y_true, y_pred):
        r"""
        Computes the *normalized* Discounted Cumulative Gain at k (nDCG@k) [KJ]_
        as a weighted sum of relevance labels of top *k* results of `y_pred`,
        normalized to the range [0, 1].
        More formally, the nDCG of each ranking *y* in `y_pred` with
        labels `y_true` is defined as

        .. math::

                \mathrm{nDCG}@k(y)  = \begin{cases} \frac{\mathrm{DCG}@k(y)}
                {\mathrm{IDCG}@k(y)} &\mbox{if } \mathrm{IDCG}@k(y) > 0 \\
                0 & \mbox{otherwise } \end{cases},

        where :math:`\mathrm{IDCG}@k(y)` is the maximum DCG@k value that can
        be achieved on *all* relevance labels (i.e., DCG@k of the sorted relevance labels).

        .. note::

                Sometimes IDCG is defined with respect to only
                the retrieved or recommended set of items *y*. This is not desirable
                as nDCG could be artificially inflated, e.g., by
                returning only one relevant item at the top
                and then filling up the ranking with and irrelevant items.

        Parameters
        ----------
        y_true : :class:`~rankereval.data.Labels`
                Ground truth labels, binary or numeric.
        y_pred : :class:`~rankereval.data.Rankings`
                Rankings to be evaluated. If `y_true` only contains one row,
                the labels in this row will be used for every ranking in `y_pred`.
                Otherwise, each row *i* in `y_pred` uses label row *i* in `y_true`.

        Returns
        -------
        computed_metric: ndarray, shape (n_rankings, )
                Computed metric for each ranking.

        Raises
        ------
        TypeError
                if `y_true` or `y_pred` are of incorrect type.

        Notes
        -----
        Edge cases:

        1. `y` is empty, i.e., :math:`|y|=0`. Per definition above, :math:`\mathrm{DCG}@k(y) = 0`.

        2. There are no items with relevance > 0 in `y_true`: :math:`\mathrm{DCG}@k(y) =` NaN
        to make it consistent with other metrics.

        3. There are no items with relevance > 0 in `y` up to *k*:  :math:`\mathrm{nDCG}@k(y) = 0`.

        Examples
        --------
        >>> from rankereval import NumericLabels, Rankings, NDCG
        >>> # use separate labels for each ranking
        >>> y_true = NumericLabels.from_dense([[1, 2, 3], [4, 5]])
        >>> y_pred = Rankings.from_ranked_indices([[0,2,1], [1,0]])
        >>> NDCG(3).score(y_true, y_pred) # doctest: +NORMALIZE_WHITESPACE
        array([0.81749351, 1.        ])

        """
        return super().score(y_true, y_pred)



"""rankerEval - A fast numpy-based implementation of ranking metrics for information retrieval and recommendation."""

__version__ = '0.2.0'
__author__ = 'Tobias Schnabel <tobias.schnabel@microsoft.com>'
__all__ = [
    'BinaryLabels',
    'NumericLabels',
    'Rankings',
    'Precision',
    'Recall',
    'F1',
    'HitRate',
    'AP',
    'ReciprocalRank',
    'DCG',
    'NDCG',
    'MeanRanks',
    'FirstRelevantRank']

from .data import BinaryLabels, NumericLabels, Rankings, DenseRankings
from .metrics import Precision, Recall, F1, HitRate, AP, ReciprocalRank, DCG, NDCG, MeanRanks, FirstRelevantRank



import io
import os

from setuptools import find_packages
from setuptools import setup


def read(filename):
    filename = os.path.join(os.path.dirname(__file__), filename)
    with io.open(filename, mode="r", encoding='utf-8') as fd:
        return fd.read()


setup(
    name="rankerEval",
    version="0.2.0",
    url="https://github.com/microsoft/rankerEval",
    license='MIT',

    author="Tobias Schnabel",
    author_email="tobias.schnabel@microsoft.com",

    description="A fast implementation of ranking metrics for information retrieval and recommendation.",
    keywords='ranking recommendation evaluation map mrr',
    long_description=read("README.rst"),
    long_description_content_type="text/x-rst",
    packages=find_packages(exclude=('tests', 'docs', 'test')),

    install_requires=['numpy>=1.18', 'numba>=0.53.0', 'scipy>=1.0'],
    python_requires='>=3.6',

    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8'
    ],
)
